/*************************************************************************
Title:    Interrupt UART library with receive/transmit circular buffers
Author:   Peter Fleury <pfleury@gmx.ch>   http://jump.to/fleury
File:     $Id: uart.c,v 1.5.2.6 2005/02/16 19:04:46 Peter Exp $
Software: AVR-GCC 3.3 
Hardware: any AVR with built-in UART, 
          tested on AT90S8515 at 4 Mhz and ATmega at 1Mhz

DESCRIPTION:
    An interrupt is generated when the UART has finished transmitting or
    receiving a byte. The interrupt handling routines use circular buffers
    for buffering received and transmitted data.
    
    The UART_RX_BUFFER_SIZE and UART_TX_BUFFER_SIZE variables define
    the buffer size in bytes. Note that these variables must be a 
    power of 2.
    
USAGE:
    Refere to the header file uart.h for a description of the routines. 
    See also example test_uart.c.

NOTES:
    Based on Atmel Application Note AVR306
                    
*************************************************************************/
#include "p24fj64ga002.h"
#include "uart.h"
#include "def.h"

/*
 *  constants and macros
 */

static volatile unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
static volatile unsigned int UART_RxHead;
static volatile unsigned int UART_RxTail;
static volatile unsigned char UART_LastRxError;
static volatile unsigned int UART_RxCnt;

static volatile unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile unsigned char UART_TxHead;
static volatile unsigned char UART_TxTail;

/*************************************************************************
Function: uart_init()
Purpose:  initialize UART and set baudrate
Input:    baudrate using macro UART_BAUD_SELECT()
Returns:  none
**************************************************************************/
void uart_init(unsigned int baudrate)
{
  unsigned int BRGVAL = 0;
	
    UART_RxHead = 0;
    UART_RxTail = 0;
	UART_RxCnt=0;
    UART_TxTail = 0;
    UART_TxHead = 0;
    
    /*
#ifdef USE_CTS
	CTS=0;
#endif
*/

	// Set baudrate
	switch (baudrate)
	{
	case BR_4800:
		BRGVAL = ((FCY)/(4800UL*16UL)-1);
		break;
	case BR_9600:
		BRGVAL = ((FCY)/(9600UL*16UL)-1);
		break;
	case BR_19200:
		BRGVAL = ((FCY)/(19200UL*16UL)-1);
		break;
	case BR_38400:
		BRGVAL = ((FCY)/(38400UL*16UL)-1);
		break;
	case BR_57600:
		BRGVAL = ((FCY)/(57600UL*16UL)-1);
		break;
	case BR_115200:
		BRGVAL = ((FCY)/(115200UL*16UL)-1);
		break;
	default:
		BRGVAL = ((FCY)/(9600UL*16UL)-1);
		break;
	}	
	
	// CHECK UART SELECTION
	//*********************
	// Set Mode ------------------------------
	U1MODEbits.UARTEN	= 1;	// UARTx is enabled
	U1MODEbits.USIDL    = 0;	// Continue module operation in Idle mode
	U1MODEbits.IREN		= 0;	// IrDA encoder and decoder disabled
	U1MODEbits.RTSMD	= 1;	// ? UxRTS pin in Simplex mode
	U1MODEbits.UEN0		= 0;	// RXD&TXD durch Uart controlliert...
	U1MODEbits.UEN1		= 0;	// RTS & CTS fr Ports freigegeben
	U1MODEbits.WAKE		= 0;	// Kein Wakeup whrend sleep
	U1MODEbits.LPBACK	= 0;	// Loopback mode is disabled
	U1MODEbits.ABAUD	= 0;	// Kein Auto Baudrat detect
	U1MODEbits.RXINV	= 0;	// UxRX Idle state is '0'
	U1MODEbits.BRGH		= 0;	// High Baud Rate Enable bit (Standard Mode, slow) 	
	U1MODEbits.PDSEL0	= 0;	// 8 Datenbits, no parity
	U1MODEbits.PDSEL1	= 0;	// 8 Datenbits, no parity
	U1MODEbits.STSEL	= 0;	// 1 Stopbit

	U1BRG	= BRGVAL;			// Baudrate Register

	// Set Status -----------------------------
	U1STAbits.UTXISEL1	= 1;	// Interrupt when a character is transferred to the Transmit
	U1STAbits.UTXISEL0	= 0;	// Shift Register (TSR) and as a result, the transmit buffer	
								// becomes empty
								
	//U1STAbits.UTXISEL1	= 0;	// Interrupt when the last character is shifted out of the 
	//U1STAbits.UTXISEL0	= 1;	// Transmit Shift Register; all transmit operations are completed
	 
	U1STAbits.UTXINV		= 0;	// UxTX Idle '0'
	U1STAbits.UTXBRK	= 0;	// Sync Break transmission disabled or completed
	U1STAbits.UTXEN		= 1;	// Transmit enabled, UxTX pin controlled by UARTx
	
	//U1STAbits.URXISEL1	= 1;	// Interrupt is set on RSR transfer, making the receive 
	//U1STAbits.URXISEL0	= 0;	// buffer 3/4 full (i.e., has 3 data characters)
	
	// Fifo ausgeschaltet
	U1STAbits.URXISEL1	= 0;	// Interrupt is set when any character is received and transferred 
	U1STAbits.URXISEL0	= 0;	// from the RSR to the receive buffer. Receive buffer has one or more characters.

	U1STAbits.ADDEN		= 0;	// Address Detect mode disabled
	
	_U1TXIP				= 0x04;	// TXD Interrupt Priority 
	_U1RXIP				= 0x04;	// RXD Interrupt Priority
	
	_U1RXIE				= 0x01;	// enable UART1 RX Interrupt 		
	_U1TXIE				= 0x01;	// enable UART1 TX Interrupt
	
}/* uart_init */
	   
unsigned int uart_getc(void)
{    
  unsigned int tmptail;
  unsigned char data;
  
    if ( UART_RxHead == UART_RxTail )
	{
		return UART_NO_DATA;   /* no data available */
    }
    
	_U1RXIE				= 0x00;	// disable UART1 RX Interrupt
	//cli();
	UART_RxCnt--;
	
//#ifdef USE_CTS
	//if (UART_RxCnt<((UART_RX_BUFFER_SIZE)/2)) CTS=0;		
//#endif

	_U1RXIE				= 0x01;	// enable UART1 RX Interrupt
	//sei();

    /* calculate /store buffer index */
    tmptail = (UART_RxTail + 1);
    
	if (tmptail>=UART_RX_BUFFER_SIZE)
		tmptail=0;

    UART_RxTail = tmptail; 
    /* get data from receive buffer */
    data = UART_RxBuf[tmptail];
    return (UART_LastRxError << 8) + data;
}/* uart_getc */


/*************************************************************************
Function: uart_putc()
Purpose:  write byte to ringbuffer for transmitting via UART
Input:    byte to be transmitted
Returns:  none          
**************************************************************************/
void UART_PutChar(unsigned char data)
{
	// Warten, bis TX Buffer frei ist	
	while (U1STAbits.UTXBF == 1);
	
	/* send one byte from buffer */
	U1TXREG = data;  
}

/*************************************************************************
Function: UART_PutChar()
Purpose:  write byte to ringbuffer for transmitting via UART
Input:    byte to be transmitted
Returns:  none          
**************************************************************************/
/*
void UART_PutChar(unsigned char data)
{
	_U1TXIE	= 0x00;	// TX Interrupt sperren
	
	// Warten, bis TX Buffer frei ist	
	while (U1STAbits.UTXBF == 1);
	
	// send one byte from buffer
	U1TXREG = data;  
	
	// Warten, bis TX Buffer frei ist	
	while (U1STAbits.UTXBF == 1);
	
	_U1TXIE	= 0x01;	// TX Interrupt freigeben
}
*/

/*************************************************************************
Function: uart_putc()
Purpose:  write byte to ringbuffer for transmitting via UART
Input:    byte to be transmitted
Returns:  none          
**************************************************************************/
void uart_putc(unsigned char data)
{

  unsigned char tmphead;
  
    tmphead  = (UART_TxHead + 1);
    
	if (tmphead>=UART_TX_BUFFER_SIZE)
		tmphead=0;
		
    while ( tmphead == UART_TxTail ); // wait for free space in buffer   
    
    UART_TxBuf[tmphead] = data;
    UART_TxHead = tmphead;
    _U1RXIE				= 0x00;	//
 
}/* uart_putc */

/*************************************************************************
Function: uart_puts()
Purpose:  transmit string to UART
Input:    string to be transmitted
Returns:  none          
**************************************************************************/
void uart_puts(const char *s )
{
    while (*s) 
      uart_putc(*s++);

}/* uart_puts */

/*************************************************************************
Function: uart_getchar()
Purpose:  Meldet die Anzahl vorhandener Bytes im RX Buffer zurck
**************************************************************************/
unsigned char uart_getchar(void)
{

  unsigned int tmptail;

    while(UART_RxHead == UART_RxTail ); 
    
    tmptail = (UART_RxTail + 1);     // calculate /store buffer index
    
	if (tmptail>=UART_RX_BUFFER_SIZE)
		tmptail=0;

    UART_RxTail = tmptail; 

	_U1RXIE	= 0x01;	// disable UART1 RX Interrupt
	//cli();
	UART_RxCnt--;

/*
#ifdef USE_CTS
	if (UART_RxCnt<((UART_RX_BUFFER_SIZE)/2))
		CTS=0;
#endif
*/
	_U1RXIE	= 0x01;	// enable UART1 RX Interrupt
	//sei();

    return UART_RxBuf[tmptail];		// get data from receive buffer
}

/*************************************************************************
Function: uart_data()
Purpose:  Meldet die Anzahl vorhandener Bytes im RX Buffer zurck
**************************************************************************/
unsigned int uart_data(void)
{
	return UART_RxCnt;	
}

/*************************************************************************
Function: UART Receive Complete interrupt
Purpose:  called when the UART has received a character
**************************************************************************/
void __attribute__((__interrupt__, auto_psv)) _U1RXInterrupt(void)
{
  unsigned int tmphead;
  unsigned char data;
   
	/* read UART status register and UART data register */ 
	data = U1RXREG;

    /* calculate buffer index */ 
    tmphead = ( UART_RxHead + 1);
	if (tmphead >= UART_RX_BUFFER_SIZE) tmphead=0;
		
	if(tmphead == UART_RxTail)
	{	    /* error: receive buffer overflow */
         UART_LastRxError = UART_BUFFER_OVERFLOW >> 8;
    }
	else
	{   /* store new index */
		UART_RxHead = tmphead;
        /* store received data in buffer */
		UART_RxBuf[tmphead] = data;
		UART_RxCnt++;
/*
#ifdef USE_CTS
		if (UART_RxCnt>((UART_RX_BUFFER_SIZE)/2))
			CTS=1;
#endif
*/

    } 
    
    _U1RXIF = 0;	// IRQ Flag zurcksetzen
}

/*************************************************************************
Function: UART Data Register Empty interrupt
Purpose:  called when the UART is ready to transmit the next byte
**************************************************************************/
void __attribute__((__interrupt__, auto_psv)) _U1TXInterrupt(void)
{	
  unsigned char tmptail;
  
    if ( UART_TxHead != UART_TxTail) 
	{
        /* calculate and store new buffer index */
        tmptail = (UART_TxTail + 1);
        
		if (tmptail>=UART_TX_BUFFER_SIZE) tmptail=0;
		
        UART_TxTail = tmptail;
        
        // get one byte from buffer and write it to UART
		while (U1STAbits.UTXBF == 1); // Warten, bis TX Buffer frei ist
	
		/* send one byte from buffer */
		U1TXREG = UART_TxBuf[tmptail];  /* start transmission */
	
    }
    else
    {
        /* tx buffer empty, disable UDRE interrupt */
        //_U1RXIE				= 0x00;	// Test
    }
    
    _U1TXIF = 0;	// IRQ Flag zurcksetzen
}





